#!/usr/bin/python
# -*- coding: ascii -*-
###########################################################################
# CTRCipher.py - Make PyCrypto CTR-mode work properly as a stream cipher
#
# Copyright (C) 2007, 2008 Dwayne C. Litzenberger <dlitz@dlitz.net>
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies and that
# both that copyright notice and this permission notice appear in
# supporting documentation.
#
# THE AUTHOR PROVIDES THIS SOFTWARE ``AS IS'' AND ANY EXPRESSED OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Country of origin: Canada
#
###########################################################################
# History:
#
#  2007-07-28 Dwayne C. Litzenberger <dlitz@dlitz.net>
#   - Pre-release (v0.0)
#
#  2008-06-17 Dwayne C. Litzenberger <dlitz@dlitz.net>
#   - Pre-release (v0.0.1)
#   - Add backward-compatibility support for Python 2.2 and 2.3
#   - Add test vectors
#
###########################################################################

from struct import pack

class CTRCounter(object):
    def __init__(self, nonce):
        if len(nonce) != 8:
            raise ValueError("nonce must be 8 bytes")
        self.nonce = nonce
        self.counter = 0L
        self._ctr_mask = ((1L << 64) - 1)
    def __call__(self):
        self.counter = (self.counter + 1) & self._ctr_mask
        if self.counter < 1:
            raise OverflowError("Counter overflowed")
        return self.nonce + pack("!Q", self.counter)

    def setNext(self, counter):
        self.counter = counter - 1

class CTRCipher(object):

    block_size = 0      # behaves like a stream cipher

    def __init__(self, key, nonce, ciphermodule):
        self._counter = CTRCounter(nonce)
        self._cipher = ciphermodule.new(key, ciphermodule.MODE_CTR, counter=self._counter)
        self._offset = 0

    def encrypt(self, plaintext):
        assert(self._offset >= 0)
        if len(plaintext) == 0:
            return ""
        if self._offset != 0:
            self._counter.setNext(self._counter.counter)
        beforePad = self._offset + len(plaintext)
        pad_size = self._cipher.block_size - (self._offset + len(plaintext)) % self._cipher.block_size
        if pad_size == self._cipher.block_size:
            pad_size = 0
        ciphertext = self._cipher.encrypt("\0" * self._offset + plaintext + "\0" * pad_size)
        ciphertext = ciphertext[self._offset:beforePad]
        self._offset = beforePad % self._cipher.block_size
        return ciphertext

    def decrypt(self, ciphertext):
        return self.encrypt(ciphertext)

    def setCounter(self, counter):
        self._counter.setNext(counter)

def test_ctrcipher():
    """Module self-test"""

    from binascii import a2b_hex

    #
    # CTRCounter tests
    #

    cc = CTRCounter("\xff\xee\xdd\xcc\xbb\xaa\x99\x88")

    # Test 1: Initial value
    result = cc()
    expected = a2b_hex("ffeeddccbbaa99880000000000000001")
    if result != expected:
        raise RuntimeError("self-test failed")

    # Test 2: Next value
    result = cc()
    expected = a2b_hex("ffeeddccbbaa99880000000000000002")
    if result != expected:
        raise RuntimeError("self-test failed")

    # Test 3: Large counter
    cc.setNext(0xFFFFFFFFfffffffeL)
    result = cc()
    expected = a2b_hex("ffeeddccbbaa9988fffffffffffffffe")
    if result != expected:
        raise RuntimeError("self-test failed")

    # Test 4: Large counter
    result = cc()
    expected = a2b_hex("ffeeddccbbaa9988ffffffffffffffff")
    if result != expected:
        raise RuntimeError("self-test failed")

    # Test 5: Counter overflow
    try:
        cc()
    except OverflowError:
        pass
    else:
        raise RuntimeError("self-test failed")

    # Test 6: _ctr_mask integer overflow
    if cc._ctr_mask != 0xFFFFFFFFffffffffL or cc._ctr_mask < 1:
        raise RuntimeError("self-test failed")

    #
    # CTRCipher tests
    # (NIST SP 800-38A test vectors)
    #
    from Crypto.Cipher import AES

    # F.5.5 CTR-AES192.decrypt
    cc = CTRCipher(
        a2b_hex("8e73b0f7da0e6452c810f32b809079e562f8ead2522c6b7b"),    # key
        a2b_hex("f0f1f2f3f4f5f6f7"),  # nonce
        AES)
    cc.setCounter(0xf8f9fafbfcfdfeffL)

    result = cc.decrypt(a2b_hex("1abc932417521ca24f2b0459fe7e6e0b"))
    expected = a2b_hex("6bc1bee22e409f96e93d7e117393172a")
    if result != expected:
        raise RuntimeError("self-test failed")

    result = cc.decrypt(a2b_hex("090339ec0aa6faefd5ccc2c6f4ce8e94"))
    expected = a2b_hex("ae2d8a571e03ac9c9eb76fac45af8e51")
    if result != expected:
        raise RuntimeError("self-test failed")

    result = cc.decrypt(a2b_hex("1e36b26bd1ebc670d1bd1d665620abf7"))
    expected = a2b_hex("30c81c46a35ce411e5fbc1191a0a52ef")
    if result != expected:
        raise RuntimeError("self-test failed")

    result = cc.decrypt(a2b_hex("4f78a7f6d29809585a97daec58c6b050"))
    expected = a2b_hex("f69f2445df4f9b17ad2b417be66c3710")
    if result != expected:
        raise RuntimeError("self-test failed")

    # F.5.5 CTR-AES256.Encrypt
    cc = CTRCipher(
        a2b_hex("603deb1015ca71be2b73aef0857d7781" + "1f352c073b6108d72d9810a30914dff4"),    # key
        a2b_hex("f0f1f2f3f4f5f6f7"),  # nonce
        AES)
    cc.setCounter(0xf8f9fafbfcfdfeffL)

    result = cc.encrypt(a2b_hex("6bc1bee22e409f96e93d7e117393172a"))
    expected = a2b_hex("601ec313775789a5b7a7f504bbf3d228")
    if result != expected:
        raise RuntimeError("self-test failed")

    result = cc.encrypt(a2b_hex("ae2d8a571e03ac9c9eb76fac45af8e51"))
    expected = a2b_hex("f443e3ca4d62b59aca84e990cacaf5c5")
    if result != expected:
        raise RuntimeError("self-test failed")

    result = cc.encrypt(a2b_hex("30c81c46a35ce411e5fbc1191a0a52ef"))
    expected = a2b_hex("2b0930daa23de94ce87017ba2d84988d")
    if result != expected:
        raise RuntimeError("self-test failed")

    result = cc.encrypt(a2b_hex("f69f2445df4f9b17ad2b417be66c3710"))
    expected = a2b_hex("dfc9c58db67aada613c2dd08457941a6")
    if result != expected:
        raise RuntimeError("self-test failed")

if __name__ == '__main__':
    test_ctrcipher()

# vim:set ts=4 sw=4 sts=4 expandtab:
